import java.util.*;
import java.io.*;

/**
 * <p>Title: Action Calculation</p>
 * <p>Description: Calculation of agents action</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: KIAM RAS</p>
 * @author Mikhail Burtsev
 * @version 1.0
 */

public class PhenoCalc {

  public int[][] popPhenotype;
  public int[] curSpecies;// number of agents for each phenotype
  public int fightStat;// number of fighting agents
  public int nSpecies;//current number of ever existed species
  public int nCurSp;//current number of non zero species
  public int kinSens;// number of kin selective agents in population
  public double[] phenoVar;// variance of genomes within given phenotype
  public double averPhenoVar;//summ variance of all genotypes within their phenotypes
  public double averPhenoC;//average variance of all centroids of genotypes within every phenotype
  public double[] phenoMVar;// variance of genomes within given phenotype
  public double averPhenoMVar;//summ variance of all genotypes within their phenotypes
  public double averPhenoMC;//average variance of all centroids of genotypes within every phenotype

  public double averKinValue;//average kin treshold

  int[][] testBed;
  int situationsToTest = 6;
  int k = WorldParams.maxEnergy;
  int[][] species;
  int[][] agentPhenotype;
  boolean fullActions = false;

  // universal constructor, where
  // energyLvls - number of gradation of energy,
  // kinLvls - number of gradation of kinship
  public PhenoCalc(int energyLvls, int kinLvls, boolean fighters) {
    situationsToTest = (ActCalc.nIn - 5)*energyLvls*kinLvls;
    species = new int[100000][situationsToTest + 1];
    nSpecies = 0;
    testBed = new int[situationsToTest][ActCalc.nIn];
    for (int i = 0; i < situationsToTest; i++) {
      for (int j = 0; j < ActCalc.nIn; j++) {
        if (j == 0)
          testBed[i][j] = k;
        else
          testBed[i][j] = 0;
      }
      if (ActCalc.highPopDensity){
        testBed[i][6] = 1000;
        testBed[i][7] = 1000;
        testBed[i][8] = 1000;
      }
    }
    int sumIndx=0;
    for (int i = 0; i < energyLvls; i++) {
      for (int z = 0; z < (ActCalc.nIn - 5); z++) {
        for (int j = 0; j < kinLvls; j++) {
          if (z < 4)
            testBed[sumIndx][z + 1] = k;
          if (z > 3)
            testBed[sumIndx][z + 1] = 1000;
          testBed[sumIndx][ActCalc.nIn -
              3] = (int) i * (k - 100) / (energyLvls - 1) + 100;
          testBed[sumIndx][ActCalc.nIn - 2] = (int) k -
              testBed[sumIndx][ActCalc.nIn - 3];
          if ((z == 4)||fighters) {
            testBed[sumIndx][z + 1] = 2000;
            if (ActCalc.noMarkers) {
              testBed[sumIndx][ActCalc.nIn - 1] = 0;
              testBed[sumIndx][ActCalc.nIn - 4] = 0;
            }
            else {
              testBed[sumIndx][ActCalc.nIn - 1] =
                  (int) (j * 1000 / (kinLvls - 1) * k /
                         (2 * WorldParams.maxMarkerValue) + 100);
              testBed[sumIndx][ActCalc.nIn -
                  4] = (int) (j * 1000 / (kinLvls - 1) + 100) / 2;
            }
          }
          sumIndx++;
        }
      }
    }
    fullActions = true;
  }

  // default constructor (for kinship)
  public PhenoCalc() {
    species = new int[100000][situationsToTest + 1];
    nSpecies = 0;
    testBed = new int[situationsToTest][ActCalc.nIn];
    for (int i = 0; i < situationsToTest; i++) {
      for (int j = 0; j < ActCalc.nIn; j++) {
        if (j == 0)
          testBed[i][j] = k;
        else
          testBed[i][j] = 0;
      }
      if (ActCalc.highPopDensity) {
        testBed[i][6] = 1000;
        testBed[i][7] = 1000;
        testBed[i][8] = 1000;
      }
    }

    //setting situations to test

    // energy low see no grass kin near
    testBed[0][ActCalc.nIn - 3] = 100; //energy
    testBed[0][ActCalc.nIn - 2] = k - testBed[0][ActCalc.nIn - 3]; //reversed energy
    testBed[0][1] = 0; // grass near
    testBed[0][5] = 2000; // agents in the knot
    testBed[0][9] = 50; // distance to the mean marker
    testBed[0][ActCalc.nIn -
        1] = (int) (100 * k / (2 * WorldParams.maxMarkerValue)); // marker distance (reversed kinship)

    // energy low see no grass non kin near
    testBed[1][ActCalc.nIn - 3] = 100; //energy
    testBed[1][ActCalc.nIn - 2] = k - testBed[1][ActCalc.nIn - 3]; //reversed energy
    testBed[1][1] = 0; // grass near
    testBed[1][5] = 2000; // agents in the knot
    testBed[1][9] = 500; // distance to the mean marker
    testBed[1][ActCalc.nIn -
        1] = (int) (1000 * k / (2 * WorldParams.maxMarkerValue)); // marker distance (reversed kinship)

    // mean energy see no grass kin near
    testBed[2][ActCalc.nIn - 3] = (int) (k / 2); //energy
    testBed[2][ActCalc.nIn - 2] = k - testBed[2][ActCalc.nIn - 3]; //reversed energy
    testBed[2][1] = 0; // grass near
    testBed[2][5] = 2000; // agents in the knot
    testBed[2][9] = 50; // distance to the mean marker
    testBed[2][ActCalc.nIn -
        1] = (int) (100 * k / (2 * WorldParams.maxMarkerValue)); // marker distance (reversed kinship)

    // mean energy see no grass non kin near
    testBed[3][ActCalc.nIn - 3] = (int) (k / 2); //energy
    testBed[3][ActCalc.nIn - 2] = k - testBed[3][ActCalc.nIn - 3]; //reversed energy
    testBed[3][1] = 0; // grass near
    testBed[3][5] = 2000; // agents in the knot
    testBed[3][9] = 500; // distance to the mean marker
    testBed[3][ActCalc.nIn -
        1] = (int) (1000 * k / (2 * WorldParams.maxMarkerValue)); // marker distance (reversed kinship)

    // a lot of mean energy see no grass kin near
    testBed[4][ActCalc.nIn - 3] = k - 100; //energy
    testBed[4][ActCalc.nIn - 2] = k - testBed[4][ActCalc.nIn - 3]; //reversed energy
    testBed[4][1] = 0; // grass near
    testBed[4][5] = 2000; // agents in the knot
    testBed[4][9] = 50; // distance to the mean marker
    testBed[4][ActCalc.nIn -
        1] = (int) (100 * k / (2 * WorldParams.maxMarkerValue)); // marker distance (reversed kinship)

    // a lot of energy see no grass non kin near
    testBed[5][ActCalc.nIn - 3] = k - 100; //energy
    testBed[5][ActCalc.nIn - 2] = k - testBed[5][ActCalc.nIn - 3]; //reversed energy
    testBed[5][1] = 0; // grass near
    testBed[5][5] = 2000; // agents in the knot
    testBed[5][9] = 500; // distance to the mean marker
    testBed[5][ActCalc.nIn -
        1] = (int) (1000 * k / (2 * WorldParams.maxMarkerValue)); // marker distance (reversed kinship) */
  }

// constructor (without kinship)
  public PhenoCalc(int s) {
    situationsToTest = s;
    species = new int[100000][situationsToTest + 1];
    nSpecies = 0;
    testBed = new int[situationsToTest][ActCalc.nIn];
    for (int i = 0; i < situationsToTest; i++) {
      for (int j = 0; j < ActCalc.nIn; j++) {
        if (j == 0)
          testBed[i][j] = k;
        else
          testBed[i][j] = 0;
      }
      if (ActCalc.highPopDensity){
        testBed[i][6] = 1000;
        testBed[i][7] = 1000;
        testBed[i][8] = 1000;
      }
    }

    //setting situations to test

    // energy low
    testBed[0][ActCalc.nIn - 3] = 100; //energy
    testBed[0][ActCalc.nIn - 2] = k - testBed[0][ActCalc.nIn - 3]; //reversed energy
    testBed[0][1] = 0; // grass near
    testBed[0][5] = 2000; // agents in the knot
    testBed[0][9] = 0; // distance to the mean marker
    testBed[0][ActCalc.nIn - 1] = 0; // marker distance (reversed kinship)

    // mean energy
    testBed[1][ActCalc.nIn - 3] = (int) (k / 2); //energy
    testBed[1][ActCalc.nIn - 2] = k - testBed[1][ActCalc.nIn - 3]; //reversed energy
    testBed[1][1] = 0; // grass near
    testBed[1][5] = 2000; // agents in the knot
    testBed[1][9] = 0; // distance to the mean marker
    testBed[1][ActCalc.nIn - 1] = 0; // marker distance (reversed kinship)

    // a lot of energy
    testBed[2][ActCalc.nIn - 3] = k - 100; //energy
    testBed[2][ActCalc.nIn - 2] = k - testBed[2][ActCalc.nIn - 3]; //reversed energy
    testBed[2][1] = 0; // grass near
    testBed[2][5] = 2000; // agents in the knot
    testBed[2][9] = 0; // distance to the mean marker
    testBed[2][ActCalc.nIn - 1] = 0; // marker distance (reversed kinship)
  }


//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<         find species

  public void findSpecies(int[][] weight) {
    popPhenotype = new int[situationsToTest][ActCalc.nAct];
    curSpecies = new int[500000];
    agentPhenotype = new int[ActCalc.nAgents][situationsToTest];
    fightStat = 0;
    boolean fightSet = false;
    averKinValue = 0;

    System.out.println("Serching for species in phenotypes space...");

    int[] curAction = new int[ActCalc.nAct];
    for (int i = 0; i < (ActCalc.nAgents); i++) { //for each agent
      for (int j = 0; j < (situationsToTest); j++) { //for every situation
        int max = -1000000000;
        int maxAct = -1;
        for (int k = 0; k < ActCalc.nAct; k++) { // calc action
          curAction[k] = 0;
          for (int z = 0; z < (ActCalc.nIn); z++) {
            curAction[k] += testBed[j][z] * weight[k + z * ActCalc.nAct][i];
          }
          if ((curAction[k] > max)&&(curAction[k] != 0)) {
            max = curAction[k];
            maxAct = k;
          }
        }
        if (maxAct == -1) maxAct = 0;
        popPhenotype[j][maxAct] += 1;
        if (fullActions) agentPhenotype[i][j] = maxAct;
          else {
            if ( (maxAct < 6) && (maxAct != 2))
              agentPhenotype[i][j] = 0; //doing nothing
            if (maxAct == 2)
              agentPhenotype[i][j] = 1; //running
            if (maxAct == 6)
              agentPhenotype[i][j] = 2; //fighting

           // if (maxAct == 5) agentPhenotype[i][j] = 3; //dividing
          }
        if ((maxAct == 6)&&(!fightSet)) {
          fightStat++;
          fightSet = true;
        }
      }
      fightSet = false;
    }
    //searching for species
    kinSens = 0;
    for (int i = 0; i < ActCalc.nAgents; i++) {
      //searching for kin sensetive phenotypes
      boolean sens = false;
      for (int k = 0; k < (situationsToTest / 2); k++) {
        if (agentPhenotype[i][2 * k] != agentPhenotype[i][2 * k + 1])
          sens = true;
      }
      if (sens){
        kinSens++;
        if (ActCalc.getAverKinValue){
          int[] wght = new int[ActCalc.nAct*ActCalc.nIn];
          for (int w = 0; w < (ActCalc.nAct*ActCalc.nIn); w++){
            wght[w] = weight[w][i];
          }
          averKinValue += getKinValue(wght);
        }
      }
        // serching for species
      boolean equals = true;
      if (nSpecies > 0) {
        for (int j = 0; j < nSpecies; j++) {
          for (int k = 0; k < situationsToTest; k++) {
            if (species[j][k] != agentPhenotype[i][k]) {
              equals = false;
              break;
            }
          }
          if (equals) {
            curSpecies[j] += 1;
            break;
          }
          if (j < (nSpecies - 1))
            equals = true;
        }
        if (!equals) {
          for (int k = 0; k < situationsToTest; k++) {
            species[nSpecies][k] = agentPhenotype[i][k];
          }
          if (sens)
            species[nSpecies][situationsToTest] = 10;
          else
            species[nSpecies][situationsToTest] = -10;
          curSpecies[nSpecies] += 1;
          nSpecies += 1;
          System.out.println("Phenospec = " + nSpecies);
        }
      }
      else {
        for (int k = 0; k < situationsToTest; k++) {
          species[0][k] = agentPhenotype[i][k];
        }
        curSpecies[0] += 1;
        nSpecies += 1;
        System.out.println("Phenospec = " + nSpecies);
      }
    }

    // current number of phenotypes in population
    nCurSp = 0;
    for (int i = 0; i < nSpecies; i++){
      if (curSpecies[i] != 0){
        nCurSp++;
      }
    }
   averKinValue = averKinValue / kinSens;

  }


//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  calculate variance of phenotypes

  public void findPhenoVar(int[][] weight, int[][] marker) {
    int[][] meanW = new int[nSpecies][ActCalc.nIn * ActCalc.nAct]; // array of mean weights for given species
    int[] meanWAver = new int[ActCalc.nIn * ActCalc.nAct]; // mean of phenotypes centroid
    int[][] meanM = new int[nSpecies][WorldParams.markerLength]; // array of mean marker for given species
    int[] meanMAver = new int[WorldParams.markerLength]; // mean of phenotypes markers centroid

    for (int i = 0; i < nSpecies; i++) {
      for (int j = 0; j < (ActCalc.nIn * ActCalc.nAct); j++) {
        meanW[i][j] = 0;
      }
      for (int j = 0; j < WorldParams.markerLength; j++) {
        meanM[i][j] = 0;
      }
    }
    phenoVar = new double[nSpecies];
    phenoMVar = new double[nSpecies];
    for (int i = 0; i < ActCalc.nAgents; i++) {
      // serching for species
      boolean equals = true;
      for (int j = 0; j < nSpecies; j++) {
        for (int k = 0; k < situationsToTest; k++) {
          if (species[j][k] != agentPhenotype[i][k]) {
            equals = false;
            break;
          }
        }
        if (equals) {
          for (int z = 0; z < (ActCalc.nIn * ActCalc.nAct); z++) {
            meanW[j][z] += weight[z][i];
          }
          for (int z = 0; z < WorldParams.markerLength; z++) {
            meanM[j][z] += marker[z][i];
          }
          break;
        }
        if (j < (nSpecies - 1))
          equals = true;
      }
    }
    for (int j = 0; j < nSpecies; j++) {
      if (curSpecies[j] > 0) {
        for (int z = 0; z < (ActCalc.nIn * ActCalc.nAct); z++) {
          meanW[j][z] = meanW[j][z] / curSpecies[j];
          meanWAver[z] += meanW[j][z];//mean of phenotypes centroid
        }
        for (int z = 0; z < WorldParams.markerLength; z++) {
          meanM[j][z] = meanM[j][z] / curSpecies[j];
          meanMAver[z] += meanM[j][z];//mean of phenotypes marker centroid
        }
      }
    }
    //calculating mean of phenotypes centroid
    for (int z = 0; z < (ActCalc.nIn * ActCalc.nAct); z++) {
      meanWAver[z] = meanWAver[z] / nSpecies;
    }
    for (int z = 0; z < WorldParams.markerLength; z++) {
      meanMAver[z] = meanMAver[z] / nSpecies;
    }
    for (int i = 0; i < ActCalc.nAgents; i++) {
      // serching for species
      boolean equals = true;
      for (int j = 0; j < nSpecies; j++) {
        for (int k = 0; k < situationsToTest; k++) {
          if (species[j][k] != agentPhenotype[i][k]) {
            equals = false;
            break;
          }
        }
        if (equals) {
          int d = 0;
          int m = 0;
          for (int z = 0; z < (ActCalc.nIn * ActCalc.nAct); z++) {
            d += (meanW[j][z] - weight[z][i]) * (meanW[j][z] - weight[z][i]);
          }
          for (int z = 0; z < WorldParams.markerLength; z++) {
            m += (meanM[j][z] - marker[z][i]) * (meanM[j][z] - marker[z][i]);
          }
          phenoVar[j] += d;
          phenoMVar[j] += m;
          break;
        }
        if (j < (nSpecies - 1))
          equals = true;
      }
    }
    averPhenoVar = 0;
    averPhenoC = 0;
    averPhenoMVar = 0;
    averPhenoMC = 0;
    for (int j = 0; j < nSpecies; j++) {
      if (curSpecies[j] > 0) {

        phenoVar[j] = phenoVar[j] / curSpecies[j];
        averPhenoVar += phenoVar[j];

        phenoMVar[j] = phenoMVar[j] / curSpecies[j];
        averPhenoMVar += phenoMVar[j];
        for (int z = 0; z < (ActCalc.nIn * ActCalc.nAct); z++) {
          averPhenoC += (meanWAver[z] - meanW[j][z])*(meanWAver[z] - meanW[j][z]);
        }
        for (int z = 0; z < WorldParams.markerLength; z++) {
          averPhenoMC += (meanMAver[z] - meanM[j][z])*(meanMAver[z] - meanM[j][z]);
        }
      }
      else{
        phenoVar[j] = 0;
        phenoMVar[j] = 0;
      }
    }
    if (nCurSp > 0){
      averPhenoVar = averPhenoVar/nCurSp; // / ActCalc.nAgents;
      averPhenoMVar = averPhenoMVar/nCurSp; // / ActCalc.nAgents;
      if (nCurSp > 1){
        averPhenoC = averPhenoC / nCurSp;
        averPhenoMC = averPhenoMC / nCurSp;
      }
      else{
        averPhenoC = 0;
        averPhenoMC = 0;
      }
    }
    else{
      averPhenoVar = 0;
      averPhenoMVar = 0;
    }
  }


//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< returns kin value for given agent

  public int getKinValue(int[] weight) {
    int kinValue = 0;
    int[] curAction = new int[ActCalc.nAct];
    int[] action = new int[situationsToTest];
    for (int kin = 1; kin < (2*WorldParams.maxMarkerValue); kin++){
      int in1 = (int) (kin * 10) / 2;
      int in2 = (int) (kin * 10 * k / (2 * WorldParams.maxMarkerValue));
      for (int j = 0; j < (situationsToTest); j++) { //for every situation
        int max = -1000000000;
        int maxAct = -1;
        for (int k = 0; k < ActCalc.nAct; k++) { // calc action
          curAction[k] = 0;
          for (int z = 0; z < (ActCalc.nIn); z++) {
            int in = testBed[j][z];
            if (z == (ActCalc.nIn - 4)) in = in1;
            if (z == (ActCalc.nIn - 1)) in = in2;
            curAction[k] += in * weight[k + z * ActCalc.nAct];
          }
          if ( (curAction[k] > max) && (curAction[k] != 0)) {
            max = curAction[k];
            maxAct = k;
          }
        }
        if (maxAct == -1)
          maxAct = 0;
        if (kin > 1) {
          if (action[j]!= maxAct) {
            kinValue = kin *10;
            break;
          }
        }
        else action[j] = maxAct;
      }
      if (kinValue > 0) break;
    }
    return kinValue;
  }

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< save species to file

  public void saveSpecies() {
    try {
      FileWriter fspecies = new FileWriter("pheno_spec.txt");
      for (int i = 0; i < nSpecies; i++) {
        fspecies.write(i + "\t");
        for (int j = 0; j < (situationsToTest + 1); j++) {
          fspecies.write(species[i][j] + "\t");
        }
        fspecies.write("\r\n");
      }
      fspecies.close();
    }
    catch (IOException e) {
      System.out.println("IO Error");
    }
  }



}
